local D = LibStub:NewLibrary("DoomCore-0.7",1)
if not D then return end
local SharedMedia = LibStub( "LibSharedMedia-3.0" )


-----------
-- Utility
-----------

local function empty(val)
  return val == nil or val == "" or (type(val) == "table" and next(val) == nil)
end
D.empty = empty


local function say(...)
  local arguments = { ... }
  local output = {}
  for i=1,#arguments do
    output[i] = tostring(arguments[i])
  end
  return table.concat(output)
end
D.say = say

local function assertType(...)
  local arguments = { ... }
  local failures = {}
  for i=1,#arguments,2 do
    local isType = type(arguments[i])
    local assertType = arguments[i+1]
    local assertTypeType = type(assertType)
    if assertTypeType == "string" then
      if isType ~= assertType then
        table.insert(failures, say("Arg #",math.ceil(i/2)," must be ",assertType,", is ",isType, "."))
      end
    elseif assertTypeType == "table"  then
      local match = false
      for _,subType in ipairs(assertType) do
        if isType == subType then
          match = true
          break
        end
      end
      if not match then
        table.insert(failures, say("Arg #",math.ceil(i/2)," must be ",table.concat(assertType,"/"),", is ",isType, "."))
      end
    end
  end
  if #failures>0 then error(table.concat(failures," "),2) end
end
D.assertType = assertType

local function capitalize(val)
  assertType(val,"string")
  return string.sub(val, 1, 1):upper() .. string.sub(val, 2):lower()
end
D.capitalize = capitalize

local function roundString(val)
  assertType(val,"number")
  if val > 1000000 then
    return math.floor(val / 10000) / 100 .. "m"
  else
    return math.floor(val / 1000) .. "k"
  end
end
D.roundString = roundString

local function orderNum(text)
  assertType(text,"string")
  text = string.upper(text)
  local order = 0
  for i=1,string.len(text) do
    order = order + (string.byte(text, i)) / math.pow(100, i)
  end
  return order
end
D.orderNum = orderNum

local function hexToCol(code)
  if empty(code) then return { r = 1, g = 1, b = 1, a = 1 } end
  return {
    r = tonumber("0x"..code:sub(1,2)) / 255,
    g = tonumber("0x"..code:sub(3,4)) / 255,
    b = tonumber("0x"..code:sub(5,6)) / 255,
    a = 1
  }
end
D.hexToCol = hexToCol

local function colToHue(col)
  assertType(col,"table")
  local hue
  local min = math.min(col.r, col.g, col.b)
  local max = math.max(col.r, col.g, col.b)
  local range = max - min
  if max == col.r then
    hue = (col.g - col.b) / range
  elseif max == col.g then
    hue = 2 + (col.b - col.r) / range
  else
    hue = 4 + (col.r - col.g) / range
  end
  hue = hue * 60
  if hue < 0 then hue = hue + 360 end
  return hue
end
D.colToHue = colToHue


-------------------
-- Object-oriented
-------------------

local function setIndex(obj, ind, noOverride)
  local meta = getmetatable(obj)
  if meta == nil then
    meta = {}
    setmetatable(obj, meta)
  end
  if noOverride and meta.__index then return end
  meta.__index = ind
end
D.setIndex = setIndex

local Inherits = {}
function Inherits:New()
  self.__index = self
  local new = {}
  setmetatable(new, self)
  new._super = self
  return new
end


------------
-- Database
------------

local Key = Inherits:New()
Key:New()
function Key:New(copyVal)
  copyVal = copyVal or {}
  assertType(copyVal,"table")
  local val = { unpack(copyVal) }
  setmetatable(val, self)
  return val
end
function Key:s(size) -- Sub-array
  assertType(size,"number")
  local newVal = Key:New{}
  for i=1,size do
    newVal[i] = self[i]
  end
  return newVal
end
function Key:r(remove) -- Remove
  assertType(remove,"number")
  return self:s(#self-remove)
end
function Key:a(...) -- Append
  local newVal = Key:New(self)
  for _,v in ipairs{...} do
    table.insert(newVal, v)
  end
  return newVal
end
D.Key = Key

local function K(val)
  return Key:New(val)
end
D.K = K

local function npairs(val, depth, ...)
  if val == nil or type(val) ~= "table" then pairs({}) end
  local i = {}
  local done = false

  if depth <= 1 then return function()
    if done then return end
    done = true
    return val
  end end

  local subPairs = npairs(val, depth - 1, select(2, ...))
  local result = {}
  local check = ...

  local function assignNext()
    local subVal = result[depth - 1]
    if subVal == nil or type(subVal) ~= "table" then
      i[depth -1], i[depth] = nil, nil
      return
    end
    while true do
      i[depth -1], i[depth] = next(subVal, i[depth -1])
      if i[depth -1]==nil then return end
      local checkType = type(check)
      if not check and i[depth -1]~=nil then return end
      if checkType == "function" and check(i[depth]) then return end
    end
  end

  return function()
    assignNext()
    while not done and i[depth] == nil do
      result = { subPairs() }
      if result[1] == nil then
        done = true
        return
      else
        i = Key:New(result):r(1)
        assignNext()
      end
    end
    if not done then return unpack(i) end
  end
end
D.npairs = npairs


-------------------
-- Crawling
-------------------

local function crawl(obj, key, noChildren, raw)
  assertType(obj,"table", key,"table")
  local max = #key
  if max == 0 then return obj elseif max < 0 then return nil end
  local node = obj
  local child
  for i=1,max do
    if raw then
      child = rawget(node, key[i])
    else
      child = node[key[i]]
    end
    if not noChildren and child == nil then
      child = {}
      node[key[i]] = child
    end
    node = child
    if node == nil then return end
  end
  return node
end
D.crawl = crawl

local NodeCrawler = Inherits:New()
function NodeCrawler:New(core)
  core = core or {}
  assertType(core,"table")
  local new = NodeCrawler._super.New(self)
  new.core = core
  return new
end
function NodeCrawler:Get(key)
  assertType(key,"table")
  return crawl(self.core, key, true)
end
function NodeCrawler:Set(key, value, noOverride)
  assertType(key,"table")
  if D.empty(key) then self.core = value end
  local parent = crawl(self.core, K(key):r(1), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local child = key[#key]
  if noOverride and parent[child] ~= nil then return end
  parent[child] = value
end
function NodeCrawler:Incr(key)
  assertType(key,"table")
  local parent = crawl(self.core, K(key):r(1), false, true)
  local child = key[#key]
  parent[child] = 1 + (parent[child] or 0)
  return parent[child]
end
D.NodeCrawler = NodeCrawler

local Configuration = NodeCrawler:New()
function Configuration:New(val)
  local new = Configuration._super.New(self, val)
  return new
end
function Configuration:AutoDefault(child)
  if not child or not child.type or not self.defaults then return end
  local default = self.defaults[child.type]
  if default then setIndex(child, default, true) end
end

function Configuration:CrawlDefaults(node, crawled)
  crawled = crawled or {}
  if not node then node = self.core end
  crawled[node] = true
  if type(node) ~= "table" then return end
  self:AutoDefault(node)
  for k,v in pairs(node) do
    if not crawled[v] then
      if type(v) == "table" and not (type(k) == "string" and k:find("_")) then
        self:CrawlDefaults(v, crawled)
      end
    end
  end
end
function Configuration:Make(parent, makeType, childName, noOverride)
  assertType(parent,"table", makeType,"string", childName,{"string","number"})
  if type(childName) == "string" then childName = childName:gsub("_","") end
  if empty(childName) then return end
  if noOverride and Configuration:Overriden(parent, childName) then return end
  local child = parent[childName]
  if child == nil then
    child = {}
    parent[childName] = child
  end
  child.type = makeType
  self:AutoDefault(child)
  return child
end
function Configuration:GetAll(parent, getType, depth)
  depth = depth or 2
  assertType(parent,"table", getType,"string", depth,"number")
  return npairs(parent, depth,
    function(val)
      return type(val) == "table" and val.type == getType
    end
  )
end
function Configuration:Delete(parent, childName)
  assertType(parent,"table", childName,{"string","number"})
  parent[childName] = nil
  parent._exclude = parent._exclude or {}
  parent._exclude[childName] = true
end
function Configuration:Excluded(parent, childName)
  assertType(parent,"table", childName,{"string","number"})
  if type(parent) ~= "table" then return end
  return parent._exclude and parent._exclude[childName]
end
function Configuration:Overriden(parent, childName)
  assertType(parent,"table", childName,{"string","number"})
  local child = parent[childName]
  if child == nil then return false end
  return child.type ~= nil
end
function Configuration:ConfGet(info)
  local val = self:Get(info)

  if type(val) == "table" and val.r and val.g and val.b then
    return val.r, val.g, val.b, val.a end
  return val
end
function Configuration:ConfGetString(info)
  return tostring(self:ConfGet(info))
end
function Configuration:ConfSet(info, r, g, b, a)
  if b then return self:Set(info, {r = r, g = g, b = b, a = a })
  elseif tonumber(r) then return self:Set(info, tonumber(r))
  else return self:Set(info, r) end
end
function Configuration:GetParent(info)
  local parent = self:Get(K(info):r(1))
  if type(parent) == "table" then return parent end
end
function Configuration:GetParentName(info)
  return info[#info - 1]
end
function Configuration:ConfGetAll(info)
  local parent = self:GetParent(info)
  if parent then return self:GetAll(parent, info[#info]) end
end
function Configuration:ConfDelete(info)
  local parent = self:GetParent(info)
  if parent then return self:Delete(parent, info[#info]) end
end
function Configuration:ConfExcluded(info)
  local parent = self:GetParent(info)
  if parent then return self:Excluded(parent, info[#info]) end
end
function Configuration:ConfOverriden(info)
  local parent = self:GetParent(info)
  if parent then return self:Overriden(parent, info[#info]) end
end
function Configuration:Number(info, val)
  return val and type(tonumber(val)) == "number"
end
function Configuration:Locked(info)
  local parent = self:GetParent(info)
  if parent then return parent.lock end
end

D.Configuration = Configuration


-----------------
-- Addon handler
-----------------

local Handler = Configuration:New()

function Handler:New(addon)
  assertType(addon,"table")
  addon.shortName = addon.shortName or addon.name
  local handler = Handler._super.New(self)
  handler.addon = addon
  setmetatable(addon,{
    __index = handler,
    __tostring = function() return addon.name end
  })
  if addon.settings then
    handler.defaults = addon.settings.defaults
    addon.settings.handler = addon
    addon.settings.handler = addon
  end
  return handler
end
function Handler:Output(...)
  if self.core._debug and self.core._debug.print then print(...) end
end
function Handler:GetVersion()
  return self.core._version
end
function Handler:SetVersion(version)
  self.core._version = version
end
function Handler:ResetDB()
  self.addon.lib.db:ResetProfile()
end
function Handler:Reset(registered)
  local addon = self.addon
  self.core = addon.lib.db.profile
  if registered ~= false then
    self:SetVersion(addon.version)
  end
  if self:ConfGetDebug{ "export" } then
    _G[addon.shortName .. "Debug"] = addon
  else
    _G[addon.shortName .. "Debug"] = nil
  end

  local settings = addon.settings
  if settings then
    settings.handler = addon
    settings.options = { type = "group", args = {} }
    if self.defaults and self.defaults.options then
      settings.options = self.defaults.options
    end
    settings.options.handler = addon
    settings.crawler = NodeCrawler:New(settings.options.args)

    if self.defaults and self.defaults.global then
      setIndex(self.core, self.defaults.global)
    end
  end
  addon:OnReset(registered)
  if addon.OnLoad then addon:OnLoad(registered) end

  if settings and settings.options then
    addon.lib.registry:RegisterOptionsTable(addon.name, settings.options)
  end
end

function Handler:MigrateDB(...)
  local keys = { ... }
  local old = Configuration:New{}
  for i,key in ipairs(keys) do
    key = K(key)
    if key == nil then key = {} end
    local subTable = self:Get(key)
    if type(subTable) ~= "table"then
      old:Set(key, subTable)
    else
      for k,v in pairs(subTable) do
        local subKey = key:a(k)
        if type(v) ~= "table" then old:Set(subKey, v) end
      end
    end
  end

  self:ResetDB()

  for i,key in ipairs(keys) do
    key = K(key)
    local subTable = old:Get(key)
    if type(subTable) ~= "table" then
      if subTable ~= self:ConfGet(key) then self:ConfSet(key, subTable) end
    else for k,v in pairs(subTable) do
      local subKey = key:a(k)
      if type(v) ~= "table" and v ~= self:ConfGet(subKey) then self:ConfSet(subKey, v) end
    end end
  end

  if self.addon.version then self:SetVersion(self.addon.version) end
end

local function debugKey(info)
  return K{ "_debug" }:a(unpack(info))
end
function Handler:ConfGetDebug(info)
  return self:ConfGet(debugKey(info))
end
function Handler:ConfSetDebug(info, r, g, b, a)
  self:ConfSet(debugKey(info), r, g, b, a)
end
function Handler:DebugExport(info, val)
  if val then
    _G[self.addon.shortName .. "Debug"] = self.addon
  else
    _G[self.addon.shortName .. "Debug"] = nil
  end
  return self:ConfSetDebug(info,val)
end
function Handler:DebugOptions(args)
  local debug = {
    type = "group",
    handler = self.addon,
    get = "ConfGetDebug",
    set = "ConfSetDebug",
    args = {
      export = {
        type = "toggle",
        name = "Export to global variable (" .. self.addon.shortName .. "Debug)",
        width = "full",
        set = "DebugExport",
        order = 0
      },
      print = {
        type = "toggle",
        name = "Print debug messages",
        width = "full",
        set = "ConfSetDebug",
        order = 1
      },
    }
  }
  for k,v in pairs(args) do
    debug.args[k] = v
  end
  return debug
end


local dbCallbacks = { "OnProfileCopied", "OnProfileDeleted", "OnProfileReset", "OnDatabaseReset", "OnProfileChanged" }
function Handler:Register(libs,defaults,defaultProfile)
  local addon = self.addon
  local lib = addon.lib or {}
  addon.lib = lib
  if not defaults and addon.settings and addon.settings.defaults and addon.settings.defaults.db then
    addon.settings.handler = addon
    defaults = addon.settings.defaults.db
  end
  defaults = defaults or {}

  lib.db = LibStub("AceDB-3.0"):New(addon.shortName .. "DB", defaults, defaultProfile)
  lib.registry = LibStub("AceConfigRegistry-3.0")
  lib.settings = LibStub("AceConfigDialog-3.0")
  lib.options = LibStub("AceDBOptions-3.0")
  for libKey, libName in pairs(libs or {}) do
    if not lib[libKey] then
      lib[libKey] = LibStub(libName,true)
    end
  end

  self:Reset(false)

  lib.settings:AddToBlizOptions(addon.name)
  local profileSettings = LibStub("AceDBOptions-3.0"):GetOptionsTable(lib.db)
  lib.registry:RegisterOptionsTable(addon.name .. " Profile", profileSettings)
  lib.settings:AddToBlizOptions(addon.name .. " Profile", "Profiles", addon.name)

  for _,callback in ipairs(dbCallbacks) do
    lib.db.RegisterCallback(self, callback, "Reset")
  end

  if addon.settings and addon.settings.debug then
    local debug = self:DebugOptions(addon.settings.debug)
    lib.registry:RegisterOptionsTable(addon.name .. " Extras", debug)
    lib.settings:AddToBlizOptions(addon.name .. " Extras","Extras", addon.name)
  end
end
D.Handler = Handler


-----------
-- Anchors
-----------

local anchors = {
  center      = "Center",
  topleft     = "Top Left",
  top         = "Top",
  topright    = "Top Right",
  left        = "Left",
  right       = "Right",
  bottomleft  = "Bottom Left",
  bottom      = "Bottom",
  bottomright = "Bottom Right"
}
D.anchors = anchors

local VERTICAL = 1
D.VERTICAL = VERTICAL
local HORIZONTAL = 2
D.HORIZONTAL = HORIZONTAL
local direction = {
  TOP    = VERTICAL,    top    = VERTICAL,
  BOTTOM = VERTICAL,    bottom = VERTICAL,
  UP     = VERTICAL,    up     = VERTICAL,
  DOWN   = VERTICAL,    down   = VERTICAL,
  LEFT   = HORIZONTAL,  left   = HORIZONTAL,
  RIGHT  = HORIZONTAL,  right  = HORIZONTAL
}
D.direction = direction
local function compound(anchor1, anchor2)
  if not anchor2 then return anchor1 end
  if not anchor1 then return anchor2 end
  if direction[anchor1] < direction[anchor2] then
    return anchor1 .. anchor2
  else
    return anchor2 .. anchor1
  end
end
D.compound = compound

local grow = { left = "Left", right = "Right", up = "Up", down = "Down" }
D.grow = grow

local growAnchors = {
  left  = { "RIGHT",  "LEFT",  -1,   0 },
  right = { "LEFT",   "RIGHT",  1,   0 },
  up    = { "BOTTOM", "TOP",    0,   1 },
  down  = { "TOP",    "BOTTOM", 0,  -1 },
  blank = { "CENTER", "CENTER", 0,   0 }
}
D.growAnchors = growAnchors
for _,dir1 in ipairs{ "left", "right" } do for _,dir2 in ipairs{ "up", "down" } do
  growAnchors[compound(growAnchors[dir1][1],growAnchors[dir2][1])] = { nil, nil, growAnchors[dir1][3], growAnchors[dir2][4] }
end end


local function makeGrid(conf)
  conf.padding = conf.padding or 0
  conf.spacing = conf.spacing or 0

  if not conf.grow2 then
    if direction[conf.grow] == HORIZONTAL then
      conf.grow2 = conf.rowGrowth
    else
      conf.grow2 = conf.columnGrowth
    end
  end

  local a1 = growAnchors[conf.grow] or growAnchors.blank
  local a2 = growAnchors[conf.grow2] or growAnchors.blank
  local comp = compound(a1[1], a2[1])
  local growComp = growAnchors[comp]

  conf.frame:ClearAllPoints()
  conf.frame:SetPoint(comp, conf.parent, conf.anchor,
    conf.x - (conf.size + conf.padding * 2) * growComp[3] / 2,
    conf.y - (conf.size + conf.padding * 2) * growComp[4] / 2)

  if not conf.els or #conf.els == 0 then
    conf.frame:SetSize(conf.size + conf.padding * 2, conf.size + conf.padding * 2)
    return
  end

  conf.els[1]:ClearAllPoints()
  conf.els[1]:SetPoint(comp, conf.frame, comp, growComp[3] * conf.padding, growComp[4] * conf.padding)

  local max = #conf.els
  if conf.max and conf.max > 0 then max = math.min(max, conf.max) end

  for i=2,max do
    conf.els[i]:ClearAllPoints()

    if conf.limit and conf.limit > 0 and (i-1) % conf.limit == 0 then
      conf.els[i]:SetPoint(a2[1], conf.els[i-conf.limit], a2[2], a2[3] * conf.spacing, a2[4] * conf.spacing)
    else
      conf.els[i]:SetPoint(a1[1], conf.els[i-1], a1[2], a1[3] * conf.spacing, a1[4] * conf.spacing)
    end
  end
  local width, height
  if conf.limit and conf.limit > 0 and max > conf.limit then
    width = conf.padding * 2 + conf.size * conf.limit + conf.spacing * (conf.limit - 1)
    local subs = math.ceil(max/conf.limit)
    height = conf.padding * 2 + conf.size * subs + conf.spacing * (subs - 1)
  else
    width = conf.padding * 2 + conf.size * max + conf.spacing * (max - 1)
    height = conf.padding * 2 + conf.size
  end
  if direction[conf.grow] == VERTICAL then
    local buf = width
    width = height
    height = buf
  end
  conf.frame:SetSize(width, height)
  local bg = {}
  if conf.background and conf.background ~= "None" then
    bg.bgFile = SharedMedia:Fetch(SharedMedia.MediaType.BACKGROUND, conf.background)
  end
  if conf.border and conf.border ~= "None" then
    bg.edgeFile = SharedMedia:Fetch(SharedMedia.MediaType.BORDER, conf.border)
    bg.edgeSize = conf.edge or 4
  end
  conf.frame:SetBackdrop(bg)
  conf.frame:Show()
end
D.makeGrid = makeGrid